/*
 * Copyright 2015 - 2023 gary@drinkingtea.net
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 */

#pragma once

#ifdef OX_USE_STDLIB
#include <string_view>
#endif

#include "bit.hpp"
#include "concepts.hpp"
#include "iterator.hpp"
#include "stringliteral.hpp"
#include "strops.hpp"
#include "vector.hpp"

namespace ox {

template<std::size_t buffLen>
class BString;

template<std::size_t buffLen>
class BasicString;

class StringView: public detail::BaseStringView {
	public:
		constexpr StringView() noexcept = default;

		constexpr StringView(const StringView &sv) noexcept = default;

#ifdef OX_USE_STDLIB
		constexpr StringView(const std::string_view &sv) noexcept: BaseStringView(sv.data(), sv.size()) {}
#endif

		template<std::size_t SmallStrSz>
		constexpr StringView(BaseStringView const&str) noexcept: BaseStringView(str.data(), str.bytes()) {}

		template<std::size_t SmallStrSz>
		constexpr StringView(const BasicString<SmallStrSz> &str) noexcept: BaseStringView(str.data(), str.len()) {}

		template<std::size_t SmallStrSz>
		constexpr StringView(const BString<SmallStrSz> &str) noexcept: BaseStringView(str.data(), str.len()) {}

		constexpr StringView(std::nullptr_t) noexcept {}

		constexpr StringView(const char *str) noexcept: BaseStringView(str) {}

		constexpr StringView(const char *str, std::size_t len) noexcept: BaseStringView(str, len) {}

		constexpr auto &operator=(StringView const&other) noexcept {
			if (&other != this) {
				set(other.data(), other.len());
			}
			return *this;
		}

};

using CRStringView = const StringView&;

constexpr auto operator==(CRStringView s1, CRStringView s2) noexcept {
	if (s2.len() != s1.len()) {
		return false;
	}
	return ox_strncmp(s1.data(), s2.data(), s1.len()) == 0;
}

constexpr auto operator<=>(CRStringView s1, CRStringView s2) noexcept {
	const auto maxLen = ox::min(s1.len(), s2.len());
	const auto a = &s1.front();
	const auto b = &s2.front();
	for (std::size_t i = 0; i < maxLen && (a[i] || b[i]); ++i) {
		if (a[i] < b[i]) {
			return -1;
		} else if (a[i] > b[i]) {
			return 1;
		}
	}
	if (s1.len() > s2.len()) {
		return 1;
	} else if (s1.len() < s2.len()) {
		return -1;
	} else {
		return 0;
	}
}

[[nodiscard]]
constexpr bool beginsWith(CRStringView base, CRStringView beginning) noexcept {
	const auto beginningLen = ox::min(beginning.len(), base.len());
	return base.len() >= beginning.len() && ox_strncmp(base.data(), beginning, beginningLen) == 0;
}

[[nodiscard]]
constexpr bool endsWith(CRStringView base, CRStringView ending) noexcept {
	const auto endingLen = ending.len();
	return base.len() >= endingLen && ox_strcmp(base.data() + (base.len() - endingLen), ending) == 0;
}

constexpr std::size_t find(CRStringView str, char search) noexcept {
	std::size_t i = 0;
	for (; i < str.len(); ++i) {
		if (str[i] == search) {
			break;
		}
	}
	return i;
}

constexpr std::size_t find(CRStringView str, CRStringView search) noexcept {
	std::size_t i = 0;
	for (; i < str.len(); ++i) {
		if (beginsWith(substr(str, i), search)) {
			break;
		}
	}
	return i;
}

template<std::size_t smallSz = 0>
constexpr ox::Vector<ox::StringView, smallSz> split(CRStringView str, char del) noexcept {
	ox::Vector<ox::StringView, smallSz> out;
	constexpr auto nextSeg = [](CRStringView current, char del) {
		return substr(current, find(current, del) + 1);
	};
	for (auto current = str; current.len(); current = nextSeg(current, del)) {
		const auto next = find(current, del);
		if (const auto s = substr(current, 0, next); s.len()) {
			out.emplace_back(s);
		}
		current = substr(current, next);
	}
	return out;
}

template<std::size_t smallSz = 0>
constexpr ox::Vector<ox::StringView, smallSz> split(CRStringView str, CRStringView del) noexcept {
	ox::Vector<ox::StringView, smallSz> out;
	constexpr auto nextSeg = [](CRStringView current, CRStringView del) {
		return substr(current, find(current, del) + del.len());
	};
	for (auto current = str; current.len(); current = nextSeg(current, del)) {
		const auto next = find(current, del);
		if (const auto s = substr(current, 0, next); s.len()) {
			out.emplace_back(s);
		}
		current = substr(current, next);
	}
	return out;
}

constexpr auto write(Writer_c auto &writer, ox::CRStringView sv) noexcept {
	return writer.write(sv.data(), sv.bytes());
}

#ifdef OX_USE_STDLIB
constexpr auto toStdStringView(CRStringView sv) noexcept {
	return std::string_view(sv.data(), sv.bytes());
}
#endif

}

constexpr ox::Result<int> ox_atoi(ox::CRStringView str) noexcept {
	int total = 0;
	int multiplier = 1;
	for (auto i = static_cast<int64_t>(str.len()) - 1; i != -1; --i) {
		auto s = static_cast<std::size_t>(i);
		if (str[s] >= '0' && str[s] <= '9') {
			total += (str[s] - '0') * multiplier;
			multiplier *= 10;
		} else {
			return OxError(1);
		}
	}
	return total;
}

